#include "DetectLineCUDA.h"
#include "cuda_runtime.h"

#include <opencv2/imgcodecs.hpp>
#include "cuda_runtime.h"
#include <opencv2/cudacodec.hpp>
#include <opencv2/cudaimgproc.hpp>
#include <opencv2/cudaarithm.hpp>

__global__ void _padKernelKernel(float* d_Dst, float* d_Src, int fftH, int fftW, int kernelH, int kernelW, int kernelY, int kernelX) {
    const int y = blockDim.y * blockIdx.y + threadIdx.y;
    const int x = blockDim.x * blockIdx.x + threadIdx.x;
    if (y < kernelH && x < kernelW)
    {
        int ky = y - kernelY;
        if (ky < 0)
        {
            ky += fftH;
        }
        int kx = x - kernelX;
        if (kx < 0)
        {
            kx += fftW;
        }
        d_Dst[ky * fftW + kx] = LOAD_FLOAT(y * kernelW + x);
    }
}

inline __device__ void mulAndScaleModified(const cufftComplex& a, const cufftComplex& b, const float& c, cufftComplex& d) {
    cufftComplex t = { c * (a.x * b.x - a.y * b.y), c * (a.y * b.x + a.x * b.y) };
    d = t;
}

__global__ void _modulateAndNormalizeKernel(cufftComplex* d_Dst, cufftComplex* d_DataSrc, cufftComplex* d_KernelSrc, int dataSize, float c) {
    const int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i >= dataSize) return;
    cufftComplex a = d_KernelSrc[i];
    cufftComplex b = d_DataSrc[i];
    cufftComplex d;
    mulAndScaleModified(a, b, c, d);
    d_Dst[i] = d;
}

__global__ void _cubeReductionKernel(float* d_Src, uchar* d_Dst, int fftH, int fftW, int dataH, int dataW, int depth) {
    int x = blockIdx.x * blockDim.x + threadIdx.x;
    int y = blockIdx.y * blockDim.y + threadIdx.y;

    if (x < dataW && y < dataH) {
        float max_value = 0.0f;
        float current_value = 0.0f;
        int offset = 0;
        for (int i = 0; i < depth; i++) {
            offset = i * fftH * fftW;
            current_value = d_Src[y * fftW + x + offset];
            if (current_value > max_value) {
                max_value = current_value;
            }
        }
        d_Dst[y * dataW + x] = (uchar)min(max_value, 255.0f);
    }
}

__global__ void _padDataClampToBorderKernel(float* d_Dst, float* d_Src, int fftH, int fftW, int dataH, int dataW, int kernelH, int kernelW, int kernelY, int kernelX) {
    const int y = blockDim.y * blockIdx.y + threadIdx.y;
    const int x = blockDim.x * blockIdx.x + threadIdx.x;
    const int borderH = dataH + kernelY;
    const int borderW = dataW + kernelX;

    if (y < fftH && x < fftW) {
        int dy, dx;
        if (y < dataH) {
            dy = y;
        }
        if (x < dataW) {
            dx = x;
        }
        if (y >= dataH && y < borderH) {
            dy = dataH - 1;
        }
        if (x >= dataW && x < borderW) {
            dx = dataW - 1;
        }
        if (y >= borderH) {
            dy = 0;
        }
        if (x >= borderW) {
            dx = 0;
        }
        d_Dst[y * fftW + x] = LOAD_FLOAT(dy * dataW + dx);
    }
}


void DetectLineCUDA::_padKernel(float* d_Dst, float* d_Src) {
    assert(d_Src != d_Dst);
    dim3 block(DETECT_TILE_X, DETECT_TILE_Y);
    dim3 grid(iDivUp(KernelW, DETECT_TILE_X), iDivUp(KernelH, DETECT_TILE_Y));

    cv::cuda::GpuMat raw_dst_mat(KernelH, KernelW, CV_32FC1, d_Dst);
    cv::Mat cpumat;
    raw_dst_mat.download(cpumat);
    cv::imwrite("d_PaddedKernel" + std::to_string(12121) + ".tif", cpumat);

    _padKernelKernel << <grid, block >> > (
        d_Dst,
        d_Src,
        FFTH,
        FFTW,
        KernelH,
        KernelW,
        KernelY,
        KernelX
        );
    cudaDeviceSynchronize();
}

void DetectLineCUDA::_modulateAndNormalize(cufftComplex* d_Dst, cufftComplex* d_DataSrc, cufftComplex* d_KernelSrc, int padding) {
	assert(FFTW % 2 == 0);
	const int dataSize = FFTH * (FFTW / 2 + padding);
	_modulateAndNormalizeKernel << <iDivUp(dataSize, 256), 256 >> > (
		d_Dst,
		d_DataSrc,
		d_KernelSrc,
		dataSize,
		1.0f / (float)(FFTW * FFTH)
		);
}

void DetectLineCUDA::_cubeReduction(float* d_Src, uchar* d_Dst) {
    dim3 block(DETECT_TILE_X, DETECT_TILE_Y);
    dim3 grid(iDivUp(Width, DETECT_TILE_X), iDivUp(Height, DETECT_TILE_Y));
    _cubeReductionKernel << <grid, block >> > (
        d_Src,
        d_Dst,
        FFTH,
        FFTW,
        Height,
        Width,
        angles.size()
        );
}

void DetectLineCUDA::_padDataClampToBorder(float* d_Dst, float* d_Src) {
	assert(d_Src != d_Dst);
	dim3 block(DETECT_TILE_X, DETECT_TILE_Y);
	dim3 grid(iDivUp(FFTW, DETECT_TILE_X), iDivUp(FFTH, DETECT_TILE_Y));
	_padDataClampToBorderKernel << <grid, block >> > (
		d_Dst,
		d_Src,
		FFTH,
		FFTW,
		Height,
		Width,
		KernelH,
		KernelW,
		KernelY,
		KernelX
		);
}
